from typing import Iterable

from django.db import models
from django.utils import timezone

__all__ = [
    "TimeStampedQuerySet",
]


def set_timestamp_for_each(items: Iterable, attrs: list[str]):
    now = timezone.now()

    for attr in attrs:
        for item in items:
            setattr(item, attr, now)

    return items


class TimeStampedQuerySet(models.QuerySet):
    _created_at_attr = "created_at"
    _updated_at_attr = "updated_at"

    def create(self, *, skip_timestamping=False, **kwargs):
        if not skip_timestamping:
            kwargs[self._created_at_attr] = kwargs[self._updated_at_attr] = (
                timezone.now()
            )

        return super().create(**kwargs)

    def update(self, *, skip_timestamping=False, **kwargs):
        if not skip_timestamping:
            kwargs[self._updated_at_attr] = timezone.now()

        return super().update(**kwargs)

    def bulk_create(self, objs, *args, skip_timestamping=False, **kwargs):
        if not skip_timestamping:
            set_timestamp_for_each(objs, [self._created_at_attr, self._updated_at_attr])

        return super().bulk_create(objs, *args, **kwargs)

    def abulk_create(self, objs, *args, skip_timestamping=False, **kwargs):
        if not skip_timestamping:
            set_timestamp_for_each(objs, [self._created_at_attr, self._updated_at_attr])

        return super().abulk_create(objs, *args, **kwargs)

    def bulk_update(self, objs, *args, skip_timestamping=False, **kwargs):
        if not skip_timestamping:
            set_timestamp_for_each(objs, [self._updated_at_attr])

        return super().bulk_update(objs, *args, **kwargs)

    def abulk_update(self, objs, *args, skip_timestamping=False, **kwargs):
        if not skip_timestamping:
            set_timestamp_for_each(objs, [self._updated_at_attr])

        return super().abulk_update(objs, *args, **kwargs)



from abc import ABC, abstractmethod
from functools import partial


class Lazy(partial):
    """
    Lazy callable object with partial preset args and kwargs.

    Example:
        >>> import datetime
        >>> lazy_now = Lazy(datetime.datetime.now, tz=...)
        >>> lazy_now() # datetime.datetime(2077, 1, 1, 1, 1, 1, 000000)
        >>> lazy_now() # datetime.datetime(2077, 1, 1, 1, 1, 2, 000000)
        >>> lazy_now() # datetime.datetime(2077, 1, 1, 1, 1, 3, 000000)
    """


class LazyCallable(ABC):
    def wrapped(self, *args, **kwargs):
        raise NotImplemented

    def __call__(self, **kwargs):
        return self.wrapped(**kwargs)

    def __init__(self, *args, **kwargs):
        self.wrapped = Lazy(self.wrapped, *args, **kwargs)



from abc import ABC
from typing import Type

from django.db import models


class ABCManager(ABC):
    QuerySetClass: Type[models.QuerySet] = NotImplemented

    def get_queryset(self) -> QuerySetClass:
        return self.QuerySetClass(self.model, using=self._db)



from datetime import datetime
from unittest.mock import MagicMock, patch

from common.testcases.model_fields_testcase import ModelFieldsTestCase
from django.test import TestCase
from django.utils import timezone
from test_models_timestamped.models import TimeStampedNote

from django_utk.db.fields import CreatedAtField, UpdatedAtField

rand_text = lambda: str(datetime.now())
frozen_time_1 = timezone.now()
frozen_time_2 = timezone.now()


class TimeStampedModelTestCase(ModelFieldsTestCase, TestCase):
    model = TimeStampedNote
    required_fields = {
        "created_at": {**CreatedAtField._required_kwargs},
        "updated_at": {**UpdatedAtField._required_kwargs},
    }

    @patch("django.utils.timezone.now")
    def test__fields__created_at(self, mock_now: MagicMock):
        """
        Checking is created_at works correctly
        """
        mock_now.return_value = frozen_time_1

        note_1: TimeStampedNote = TimeStampedNote.objects.create(text=rand_text())
        self.assertEqual(note_1.created_at, frozen_time_1)

        note_2: TimeStampedNote = TimeStampedNote.objects.create(text=rand_text())
        self.assertEqual(note_2.created_at, frozen_time_1)

        mock_now.return_value = frozen_time_2

        note_1.text = rand_text()
        note_1.save()
        note_1.refresh_from_db()

        self.assertEqual(note_1.created_at, frozen_time_1)
        self.assertEqual(note_2.created_at, frozen_time_1)

    @patch("django.utils.timezone.now")
    def test__fields__updated_at(self, mock_now: MagicMock):
        """
        Checking is created_at works correctly
        """
        mock_now.return_value = frozen_time_1

        note: TimeStampedNote = TimeStampedNote.objects.create(text=rand_text())
        self.assertEqual(note.updated_at, frozen_time_1)

        mock_now.return_value = frozen_time_2

        note.text = rand_text()
        note.save()
        note.refresh_from_db()

        self.assertEqual(note.updated_at, frozen_time_2)



from functools import wraps


def typehint(typed_method: callable = None, *, source: type = None):
    """
    Method decorator for conveniently adding type hints to child class methods
    without initializing the method body with `super` call and be left `pass`

    >>> from django_utk.utils.typehint import typehint
    >>>
    >>> class Calculator:
    >>>     def is_greater(self, a, b):
    >>>         return a > b
    >>>
    >>>     def is_lower(self, a, b):
    >>>         return a < b
    >>>
    >>> class TypedCalculator(Calculator):
    >>>     @typehint               # simple way
    >>>     def is_greater(self, a: int, b: int) -> bool:
    >>>         pass
    >>>
    >>>     @typehint(Calculator)   # better performance
    >>>     def is_lower(self, a: int, b: int) -> bool:
    >>>         pass
    >>>
    >>>
    >>> calc = TypedCalculator()
    >>> print(calc.is_greater(1, 0))    # True
    >>> print(calc.is_lower(1, 0))      # False
    """

    if not typed_method and not source:
        raise TypeError(
            f"{typehint.__repr__()} decorator require at least one of two arguments: 'typed_method' or 'source'"
        )

    if isinstance(typed_method, type):
        typed_method, source = None, typed_method

    def typehint_wrapper(method: callable):
        if source:
            super_proxy = wraps(method)(getattr(source, method.__name__))

        else:

            @wraps(method)
            def super_proxy(self, *args, **kwargs):
                if isinstance(self, type):
                    cls = self
                    args = (self, *args)
                else:
                    cls = type(self)

                super_obj = super(cls, self)
                super_method = getattr(super_obj, method.__name__)

                return super_method(*args, **kwargs)

        return super_proxy

    if typed_method:
        return typehint_wrapper(typed_method)
    else:
        return typehint_wrapper



from django_utk.validators.base import BaseValidator
from django_utk.validators.hashes import (
    MD5Validator,
    SHA1Validator,
    SHA256Validator,
    SHA512Validator,
)
from django_utk.validators.uuid import (
    UUID1Validator,
    UUID2Validator,
    UUID3Validator,
    UUID4Validator,
    UUID5Validator,
)



from django_utk.db.models.timestamped.fields import *
from django_utk.db.models.timestamped.fields import __all__ as fields_all
from django_utk.db.models.timestamped.models import *
from django_utk.db.models.timestamped.models import __all__ as models_all
from django_utk.db.models.timestamped.querysets import TimeStampedQuerySet

__all__ = [
    *models_all,
    *fields_all,
]
__all__ += [
    "TimeStampedQuerySet",
]



from abc import ABC, ABCMeta, abstractmethod
from typing import Callable, Dict, List, Type

from django.db import models

from django_utk.tests.faker.sequences import BaseSequence
from django_utk.utils.lazy import Lazy, LazyCallable
from django_utk.utils.popattr import popattr
from django_utk.utils.typehint import typehint


class FieldFactory:
    @classmethod
    def from_any(cls, value: any) -> "FieldFactory":
        if callable(value):
            return cls(value)
        else:
            return cls(Lazy(lambda: value))

    def __init__(self, getter: Callable):
        self.getter = getter

    def __call__(self):
        return self.getter()


class FactoryOptions:
    model: Type[models.Model]
    fields: List[str]
    fields_set: Dict[str, "FieldFactory"]
    factory: "Factory"

    @classmethod
    def from_factory(cls, factory) -> "FactoryOptions":
        return cls(getattr(factory, "Meta", None), factory)

    def __init__(self, options, factory: "Factory"):
        self.model = getattr(options, "model", NotImplemented)
        self.fields = getattr(options, "fields", None)
        self.fields_set = getattr(options, "fields_set", dict())
        self.factory = factory


class FactoryMeta(ABCMeta):
    def __new__(cls, name, bases, attrs):
        new_class: BaseFactory = super().__new__(cls, name, bases, attrs)  # noqa

        if not popattr(new_class, "__no_meta__", False):
            new_class._meta = FactoryOptions.from_factory(new_class)

            for base in reversed(bases):
                if issubclass(base, BaseFactory) and hasattr(base, "_meta"):
                    new_class._meta.fields_set.update(base._meta.fields_set)

            for attr_name, attr_value in attrs.items():

                is_attr_method = FactoryMeta.is_attr_method(attr_name, attr_value)

                if not is_attr_method:
                    new_class._meta.fields_set[attr_name] = FieldFactory.from_any(
                        attr_value
                    )

            if new_class._meta.fields is None:
                new_class._meta.fields = list(new_class._meta.fields_set.keys())
            else:
                new_class._meta.fields_set = {
                    field_name: field_factory
                    for field_name, field_factory in new_class._meta.fields_set.items()
                    if field_name in new_class._meta.fields
                }

        return new_class

    @staticmethod
    def is_attr_method(attr_name: str, attr_value: any):
        return callable(attr_value) and not isinstance(
            attr_value, (SubFactory, BaseSequence, Lazy, LazyCallable)
        )


class BaseFactory(ABC):

    @classmethod
    @abstractmethod
    def get_model(cls) -> type:
        raise NotImplemented

    @classmethod
    @abstractmethod
    def get_fields_defaults(cls) -> dict[str, Callable]:
        return NotImplemented

    @classmethod
    def get_init_values(cls, **kwargs):
        return {
            **{
                field_name: field_getter()
                for field_name, field_getter in cls.get_fields_defaults().items()
            },
            **{
                kwarg_name: FieldFactory.from_any(kwarg_value)
                for kwarg_name, kwarg_value in kwargs
            },
        }

    @classmethod
    def init_obj(cls, **kwargs):
        model = cls.get_model()
        init_values = cls.get_init_values(**kwargs)
        return model(**init_values)

    @classmethod
    def create(cls, **kwargs):
        """
        Create a single instance
        """
        return cls.init_obj(**kwargs)

    @classmethod
    def create_batch(cls, count, **kwargs) -> list:
        """
        Create many of instances
        """
        return [cls.init_obj(**kwargs) for _ in range(count)]

    def __new__(cls, **kwargs):
        return cls.create(**kwargs)

    @classmethod
    def sub_factory(cls, **kwargs):
        return SubFactory(cls, **kwargs)


class Factory(BaseFactory, metaclass=FactoryMeta):
    __no_meta__ = True
    _meta: FactoryOptions

    @classmethod
    def get_model(cls) -> type:
        return cls._meta.model

    @classmethod
    def get_fields_defaults(cls) -> dict:
        return cls._meta.fields_set

    @classmethod
    def create(cls, **kwargs):
        """
        Create a single instance
        """
        obj = super().create(**kwargs)
        obj.save()
        return obj

    @classmethod
    def create_batch(cls, count, **kwargs) -> list:
        """
        Create many of instances
        """
        objs = super().create_batch(count, **kwargs)
        return cls._meta.model.objects.bulk_create(objs)


class SubFactory(Lazy):
    @typehint
    def ___new__(self, factory: Factory, **kwargs) -> "SubFactory":
        pass



import random
from typing import Any, Callable, Iterable

from django_utk.utils.lazy import LazyCallable

__all__ = [
    "RandChoices",
    "RandChoice",
]


class RandChoices(LazyCallable):
    wrapped = random.choices

    def __init__(
        self,
        population: Iterable[Any],
        k=1,
        *,
        weights: list[int] = None,
        cum_weights: list[int] = None,
    ):
        super().__init__(
            population=population, weights=weights, cum_weights=cum_weights, k=k
        )


class RandChoice(RandChoices):
    def __init__(
        self,
        population: Iterable[Any],
        *,
        weights: list[int] = None,
        cum_weights: list[int] = None,
    ):
        super().__init__(
            population=population, weights=weights, cum_weights=cum_weights, k=1
        )

    def wrapped(self, **kwargs):
        try:
            return super().wrapped(**kwargs)[0]
        except IndexError:
            return None



import random
import string

from django_utk.utils.lazy import LazyCallable

__all__ = [
    "RandString",
]

from django_utk.utils.typehint import typehint


class RandString(LazyCallable):
    ALPHABET = string.hexdigits
    LENGTH = 32

    def wrapped(self, length: int, alphabet: str):
        return "".join(random.choices(alphabet, k=length))

    def __init__(self, length: int = LENGTH, *, alphabet: str = None):
        if alphabet is None:
            alphabet = self.ALPHABET

        super().__init__(length=length, alphabet=alphabet)

    @typehint
    def __call__(self, *args, **kwargs) -> str:
        pass



from typing import Never, NoReturn

from django.db import models

__all__ = [
    "DBViewQuerySet",
]


class DBViewQuerySet(models.QuerySet):
    @classmethod
    def raise__method_not_allowed(cls, *args, **kwargs) -> NoReturn:
        raise AssertionError(
            f"Method is not allowed, for class '{cls.__name__}' due to it's db view"
        )

    def _batched_insert(self, *args, **kwargs) -> Never:
        self.raise__method_not_allowed()

    def bulk_create(self, *args, **kwargs) -> Never:
        self.raise__method_not_allowed()

    def bulk_update(self, *args, **kwargs) -> Never:
        self.raise__method_not_allowed()

    def create(self, *args, **kwargs) -> Never:
        self.raise__method_not_allowed()

    def delete(self, *args, **kwargs) -> Never:
        self.raise__method_not_allowed()

    def get_or_create(self, *args, **kwargs) -> Never:
        self.raise__method_not_allowed()

    def update(self, *args, **kwargs) -> Never:
        self.raise__method_not_allowed()

    def update_or_create(self, *args, **kwargs) -> Never:
        self.raise__method_not_allowed()



import string
from unittest import TestCase

from django_utk.tests import faker
from django_utk.tests.utils import run_100_times

rand_int = faker.RandInt(1, 42)


class RandStringTestCase(TestCase):
    @run_100_times
    def assertResultValid(self, string_factory: callable, length: int, alphabet: set):
        result = string_factory()

        self.assertEqual(len(result), length)
        self.assertEqual(len(set(result) - set(alphabet)), 0)

    def test_default(self):
        string_fct = faker.RandString()

        self.assertResultValid(
            string_fct, faker.RandString.LENGTH, faker.RandString.ALPHABET
        )

    @run_100_times
    def test__alphabet(self):
        string_len = rand_int()
        alphabet = string.printable[: min(len(string.printable), rand_int())]
        string_fct = faker.RandString(string_len, alphabet=alphabet)
        self.assertResultValid(string_fct, string_len, set(alphabet))



from django.db import models
from django.utils.translation import gettext_lazy as _

from django_utk.db.models.softdelete import SoftDeleteQuerySet
from django_utk.db.models.softdelete.managers import SoftDeleteManager
from django_utk.db.models.utils import get_model_meta as meta

__all__ = [
    "BaseSoftDeleteModel",
    "SoftDeleteModel",
]


class BaseSoftDeleteModel:
    deleted_at = models.DateTimeField(
        verbose_name=_("Deleted at"),
        default=None,
        null=True,
        blank=True,
    )


class SoftDeleteModel(BaseSoftDeleteModel, models.Model):
    deleted_at: models.DateTimeField

    objects = SoftDeleteManager()
    all_objects = SoftDeleteQuerySet.as_manager()

    class Meta:
        abstract = True

    def validate_existence(self, action: str):
        if self.pk is None:
            raise ValueError(
                "Failed to {action} object {obj_name} because it's {pk_attname} attribute is set to None".format(
                    action=action,
                    obj_name=meta(self).object_name,
                    pk_attname=meta(self).pk.attname,
                )
            )

    def delete(self, **kwargs):
        self.validate_existence(action=SoftDeleteModel.delete.__name__)
        model = type(self)
        model.all_objects.filter(pk=self.pk).delete()

    def restore(self):
        self.validate_existence(action=SoftDeleteModel.restore.__name__)
        model = type(self)
        model.all_objects.filter(pk=self.pk).restore()



from unittest import TestCase

from django_utk.tests import faker
from django_utk.tests.utils import run_10k_times


class RandNumericTestCase:
    NumericFactory = NotImplemented

    a = NotImplemented
    b = NotImplemented
    return_type = NotImplemented

    @run_10k_times
    def test__default(self):
        rand_fct = self.NumericFactory()
        value = rand_fct()

        self.assertIsInstance(value, self.return_type)
        self.assertGreaterEqual(value, self.NumericFactory.MIN)
        self.assertLessEqual(value, self.NumericFactory.MAX)

    @run_10k_times
    def test__gte(self):
        rand_fct_1 = self.NumericFactory()
        rand_fct_2 = self.NumericFactory(self.a)
        rand_fct_3 = self.NumericFactory(a=self.a)

        value_1 = rand_fct_1(a=self.a)
        value_2 = rand_fct_2()
        value_3 = rand_fct_3()

        self.assertGreaterEqual(value_1, self.a)
        self.assertGreaterEqual(value_2, self.a)
        self.assertGreaterEqual(value_3, self.a)

        self.assertLessEqual(value_1, self.NumericFactory.MAX)
        self.assertLessEqual(value_2, self.NumericFactory.MAX)
        self.assertLessEqual(value_3, self.NumericFactory.MAX)

    @run_10k_times
    def test__lte(self):
        rand_fct_1 = self.NumericFactory()
        rand_fct_2 = self.NumericFactory(b=self.a)

        value_1 = rand_fct_1(b=self.a)
        value_2 = rand_fct_2()

        self.assertGreaterEqual(value_1, self.NumericFactory.MIN)
        self.assertGreaterEqual(value_2, self.NumericFactory.MIN)

        self.assertLessEqual(value_1, self.a)
        self.assertLessEqual(value_2, self.a)

    @run_10k_times
    def test__gte_lte(self):
        rand_fct_1 = self.NumericFactory()
        rand_fct_2 = self.NumericFactory(self.a, self.b)
        rand_fct_3 = self.NumericFactory(a=self.a, b=self.b)
        rand_fct_4 = self.NumericFactory(a=self.a)
        rand_fct_5 = self.NumericFactory(b=self.b)

        value_1 = rand_fct_1(a=self.a, b=self.b)
        value_2 = rand_fct_2()
        value_3 = rand_fct_3()
        value_4 = rand_fct_4(b=self.b)
        value_5 = rand_fct_5(a=self.a)

        self.assertGreaterEqual(value_1, self.a)
        self.assertGreaterEqual(value_2, self.a)
        self.assertGreaterEqual(value_3, self.a)
        self.assertGreaterEqual(value_4, self.a)
        self.assertGreaterEqual(value_5, self.a)

        self.assertLessEqual(value_1, self.b)
        self.assertLessEqual(value_2, self.b)
        self.assertLessEqual(value_3, self.b)
        self.assertLessEqual(value_4, self.b)
        self.assertLessEqual(value_5, self.b)


class RandIntTestCase(RandNumericTestCase, TestCase):
    NumericFactory = faker.RandInt

    a = 42
    b = a + a
    return_type = int


class RandFloatTestCase(RandNumericTestCase, TestCase):
    NumericFactory = faker.RandFloat

    a = 42.42
    b = a + a
    return_type = float



from typing import Self

from django.db import models
from django.utils import timezone

__all__ = [
    "SoftDeleteQuerySet",
]


class SoftDeleteQuerySet(models.QuerySet):
    def delete(self):
        self.update(deleted_at=timezone.now())
        return True, self.count()

    def restore(self):
        self.update(deleted_at=None)
        return True, self.count()

    def hard_delete(self):
        return super(SoftDeleteQuerySet, self).delete()

    def filter(self, *args, **kwargs) -> Self:
        return super(SoftDeleteQuerySet, self).filter(*args, **kwargs)



import random
from unittest import TestCase
from unittest.mock import MagicMock

from django_utk.tests import faker


class RandChoicesTestCase(TestCase):

    A = 42
    B = 10_000

    def test__default(self):
        sequence_fct = faker.Sequence()

        for i in range(self.B):
            self.assertEqual(sequence_fct(), i)

    def test__start(self):
        sequence_fct = faker.Sequence(start=self.A)

        for i in range(self.A, self.B):
            self.assertEqual(sequence_fct(), i)

    def test__end(self):
        sequence_fct = faker.Sequence(end=self.A)

        for i in range(self.A):
            self.assertEqual(sequence_fct(), i)

        with self.assertRaises(StopIteration):
            sequence_fct()

    def test__start_end(self):
        sequence_fct = faker.Sequence(start=self.A, end=self.B)

        for i in range(self.A, self.B):
            self.assertEqual(sequence_fct(), i)

        with self.assertRaises(StopIteration):
            sequence_fct()

    def test__step(self):
        sequence_fct = faker.Sequence(step=self.A)

        for i in range(self.B):
            self.assertEqual(sequence_fct(), (i * self.A))

    def test__handler(self):
        handler = MagicMock()

        sequence_fct = faker.Sequence(handler)

        for i in range(self.B):
            self.assertEqual(handler.call_count, i)
            sequence_fct()
            self.assertEqual(handler.call_args[0][0], i)


class ForEachTestCase(TestCase):

    ITEMS_LEN = 100
    ITEMS = [i for i in range(ITEMS_LEN)]

    @classmethod
    def setUpClass(cls):
        random.shuffle(cls.ITEMS)

    def test__default(self):
        foreach_fct = faker.ForEach(self.ITEMS)

        for i in range(self.ITEMS_LEN):
            self.assertEqual(foreach_fct(), self.ITEMS[i])



from django_utk.db.models.base import *
from django_utk.db.models.base import __all__ as base_all
from django_utk.db.models.dbview import *
from django_utk.db.models.dbview import __all__ as dbview_all
from django_utk.db.models.timestamped import *
from django_utk.db.models.timestamped import __all__ as temestamped_all
from django_utk.db.models.timestamped.fields import *
from django_utk.db.models.timestamped.fields import __all__ as fields_all

__all__ = [
    *base_all,
    *dbview_all,
    *temestamped_all,
    *fields_all,
]



import sys

"""
PYTHON_VERSION
    Just shortcut to get current running python interpreter version
    
    For example "Check is python version lower 3.11":
    
    >>> from django_utk.utils.env import PYTHON_VERSION
    >>>
    >>> if PYTHON_VERSION < (3, 11):
    >>>    foo()

"""

PYTHON_VERSION = sys.version_info[:2]



from django.db.models.manager import BaseManager

from django_utk.db.models.softdelete import managers, models, querysets


class SomeQS(querysets.SoftDeleteQuerySet):
    def asdf(self):
        pass


SomeManager = BaseManager.from_queryset(SomeQS)


print(managers.SoftDeleteManager.__dict__)



from dataclasses import dataclass
from unittest import TestCase
from unittest.mock import MagicMock

from django_utk.tests import faker
from django_utk.tests.factories import Factory
from django_utk.utils.typehint import typehint

small_int = faker.RandInt(2, 42)


class FactoryTestCase(TestCase):
    def test__init__const(self):
        PersonModel = MagicMock()

        person_name = "person-name"
        person_age = 33
        person_fiends = []  # haha

        class PersonFactory(Factory):
            name = person_name
            age = person_age
            friends = person_fiends

            class Meta:
                model = PersonModel

        for i in range(small_int()):
            person: PersonModel = PersonFactory()

            model_calls = PersonModel.call_args_list
            kwargs = model_calls[i].kwargs

            self.assertIsInstance(person, MagicMock)
            self.assertEqual(len(model_calls), i + 1)
            self.assertEqual(kwargs["name"], person_name)
            self.assertEqual(kwargs["age"], person_age)
            self.assertEqual(kwargs["friends"], person_fiends)
            self.assertEqual(
                len(kwargs) - 2,  # MagicMock's internal kwargs
                len([person_name, person_age, person_fiends]),
            )

    def test__init__SubFactory(self):
        @dataclass
        class SkillModel:
            name: str

            def save(self):
                pass

        @dataclass
        class PersonModel:
            skill: SkillModel

            def save(self):
                pass

        class SkillFactory(Factory):
            name = "skill-name"

            class Meta:
                model = SkillModel
                fields = ["name"]

        class PersonFactory(Factory):
            skill = SkillFactory.sub_factory()

            class Meta:
                model = PersonModel
                fields = ["skill"]

        for i in range(small_int()):
            person = PersonFactory()

            self.assertIsInstance(person.skill, SkillModel)

    def test__init__Sequence(self):
        person_name_suffix = "-person"

        @dataclass
        class PersonModel:
            name: str

            def save(self):
                pass

        class PersonFactory(Factory):
            name = faker.Sequence(lambda n: f"{n}{person_name_suffix}")

            @typehint
            def __new__(cls, *args, **kwargs) -> PersonModel:
                pass

            class Meta:
                model = PersonModel
                fields = ["name"]

        for i in range(small_int()):
            person = PersonFactory()

            self.assertTrue(person.name.endswith(person_name_suffix))
            self.assertEqual(person.name, f"{i}{person_name_suffix}")

    def test__init__ForEach(self):
        persons_amount = small_int()
        persons_names = [f"person-name-{i}" for i in range(persons_amount)]

        @dataclass
        class PersonModel:
            name: str

            def save(self):
                pass

        class PersonFactory(Factory):
            name = faker.ForEach(persons_names)

            @typehint
            def __new__(cls, *args, **kwargs) -> PersonModel:
                pass

            class Meta:
                model = PersonModel
                fields = ["name"]

        for i in range(persons_amount):
            person = PersonFactory()

            self.assertEqual(person.name, persons_names[i])

        with self.assertRaises(StopIteration):
            PersonFactory()

    def test__init__RandData(self):
        @dataclass
        class PersonModel:
            age: int
            weight: float
            name: str

            def save(self):
                pass

        class PersonFactory(Factory):
            age = faker.RandInt()
            weight = faker.RandFloat()
            name = faker.RandString()

            @typehint
            def __new__(cls, *args, **kwargs) -> PersonModel:
                pass

            class Meta:
                model = PersonModel
                fields = ["age", "weight", "name"]

        for i in range(small_int()):
            person = PersonFactory()

            self.assertLessEqual(person.age, faker.RandInt.MAX)
            self.assertGreaterEqual(person.age, faker.RandInt.MIN)

            self.assertLessEqual(person.weight, faker.RandFloat.MAX)
            self.assertGreaterEqual(person.weight, faker.RandFloat.MIN)
            self.assertGreater(len(person.name), 0)
            self.assertTrue(
                all(letter in faker.RandString.ALPHABET for letter in person.name)
            )

    def test__init__Choice_and_Choices(self):
        persons_amount = small_int()
        persons_names = [f"person-name-{i}" for i in range(persons_amount)]

        @dataclass
        class PersonModel:
            name: str
            friends: list[str]

            def save(self):
                pass

        class PersonFactory(Factory):
            name = faker.RandChoice(persons_names)
            friends = faker.RandChoices(persons_names, k=small_int())

            @typehint
            def __new__(cls, *args, **kwargs) -> PersonModel:
                pass

            class Meta:
                model = PersonModel
                fields = ["name", "friends"]

        for i in range(small_int()):
            person = PersonFactory()

            self.assertIn(person.name, persons_names)

            for friend_name in person.friends:
                self.assertIn(friend_name, persons_names)



from typing import Type, Union

from django.db import models
from django.db.models.options import Options


def get_model_meta(model: Union[Type[models.Model], models.Model]) -> Options:
    return getattr(model, "_meta", getattr(model, "Meta"))



from django.db import models

__all__ = [
    "CreatedAtField",
    "UpdatedAtField",
]


class CreatedAtField(models.DateTimeField):
    """
    Unified 'created_at' field for timestamped models

    Have preset automatic filling by current datetime on instance creation
    and readonly mode for admin panel
    """

    _required_kwargs = {
        "auto_now_add": True,
        "blank": True,
        "editable": False,
    }

    def __init__(self, *args, **kwargs):
        kwargs.update(self._required_kwargs)
        models.DateTimeField.__init__(self, *args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs.update(self._required_kwargs)
        return name, path, args, kwargs


class UpdatedAtField(models.DateTimeField):
    """
    Unified 'created_at' field for timestamped models

    Have preset automatic filling by current datetime on instance update
    and readonly mode for admin panel
    """

    _required_kwargs = {
        "auto_now": True,
        "blank": True,
        "editable": False,
    }

    def __init__(self, *args, **kwargs):
        kwargs.update(self._required_kwargs)
        models.DateTimeField.__init__(self, *args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs.update(self._required_kwargs)
        return name, path, args, kwargs



from django.db import models

__all__ = ["DeleteUsingManager"]


class DeleteUsingManager:
    """
    Replaces default Model.delete() feature (which users Collector.delete())
    by deleting using model's manager Model.objects.delete()

    Use "_objects_attr" class property to specify model's manager
    """

    _objects_attr = "objects"

    def delete(self, **kwargs):
        objects = getattr(type(self), self._objects_attr)
        return objects.filter(pk=self.pk).delete()



from common.abc import ABCManager
from django.db import models

from django_utk.db.models.timestamped.querysets import TimeStampedQuerySet


class BaseTimeStampedManager(ABCManager):
    QuerySetClass = TimeStampedQuerySet


class TimeStampedManager(BaseTimeStampedManager, models.Manager):
    pass



INSTANCE_KEY = "__instance__"


def get_instance(cls: type, getter: callable, key: str = INSTANCE_KEY):
    instance = getattr(cls, key, None) or getter(cls)
    setattr(cls, key, instance)
    return instance


class SingletonMeta(type):
    def __call__(cls, *args, **kwargs):
        return get_instance(
            cls,
            lambda kls: super(SingletonMeta, kls).__call__(*args, **kwargs),
        )


class Singleton(metaclass=SingletonMeta):
    pass



from django.db import models

from django_utk.db.models import TimeStampedModel
from django_utk.db.models.timestamped import CreatedAtField, UpdatedAtField


class TimeStampedNote(TimeStampedModel):
    """
    Just regular note with text and timestamps
    """

    created_at: CreatedAtField
    updated_at: UpdatedAtField

    text = models.TextField()



import uuid

from django_utk.tests.faker import RandString

__all__ = [
    "UUID4",
]


class UUID4(RandString):
    def wrapped(self, length: int, alphabet: str):
        return str(uuid.uuid4().hex)

    def __init__(self):
        super().__init__()



import string
from unittest import TestCase
from unittest.mock import MagicMock

from django_utk.tests import faker
from django_utk.utils.lazy import Lazy

small_int = faker.RandInt(2, 42)
rand_str = faker.RandString(alphabet=string.ascii_letters)


class RunNTimesTestCase(TestCase):

    def test__no_args(self):
        args = []
        kwargs = {}
        original_func = MagicMock()
        lazy_func = Lazy(original_func)

        for i in range(small_int()):
            lazy_func()
            self.assertEqual(original_func.call_count, i + 1)
            self.assertListEqual(list(original_func.call_args_list[i].args), args)
            self.assertDictEqual(dict(original_func.call_args_list[i].kwargs), kwargs)

    def test__post_args(self):
        original_func = MagicMock()
        lazy_func = Lazy(original_func)

        for i in range(small_int()):
            lazy_func()
            self.assertEqual(original_func.call_count, i + 1)
            self.assertListEqual(list(original_func.call_args_list[i].args), [])
            self.assertDictEqual(dict(original_func.call_args_list[i].kwargs), {})

    def test__initial_args(self):
        initial_kwargs = {rand_str(): rand_str() for _ in range(small_int())}
        initial_args = [rand_str() for _ in range(small_int())]
        original_func = MagicMock()
        lazy_func = Lazy(original_func, *initial_args, **initial_kwargs)

        for i in range(small_int()):
            lazy_func()
            self.assertEqual(original_func.call_count, i + 1)
            self.assertListEqual(
                list(original_func.call_args_list[i].args), initial_args
            )
            self.assertDictEqual(
                dict(original_func.call_args_list[i].kwargs), initial_kwargs
            )

    def test__initial_args__and__post_args(self):
        args = [rand_str() for _ in range(small_int())]
        initial_args = [rand_str() for _ in range(small_int())]

        kwargs = {rand_str(): rand_str() for _ in range(small_int())}
        initial_kwargs = {rand_str(): rand_str() for _ in range(small_int())}

        calls_count = 0

        original_func = MagicMock()
        lazy_func = Lazy(original_func, *initial_args, **initial_kwargs)

        calls_count, _ = (calls_count + 1), lazy_func(*args, **kwargs)

        self.assertEqual(original_func.call_count, calls_count)
        self.assertListEqual(
            list(original_func.call_args_list[calls_count - 1].args),
            [*initial_args, *args],
        )
        self.assertDictEqual(
            dict(original_func.call_args_list[calls_count - 1].kwargs),
            {**kwargs, **initial_kwargs},
        )

        calls_count, _ = (calls_count + 1), lazy_func()

        self.assertEqual(original_func.call_count, calls_count)
        self.assertListEqual(
            list(original_func.call_args_list[calls_count - 1].args), initial_args
        )
        self.assertDictEqual(
            dict(original_func.call_args_list[calls_count - 1].kwargs), initial_kwargs
        )



from abc import ABC, abstractmethod
from functools import partial

from django.utils.deconstruct import deconstructible

from django_utk.validators.base import BaseValidator

__all__ = [
    "is_hash_valid",
    "is_valid_md5",
    "is_valid_sha1",
    "is_valid_sha256",
    "is_valid_sha512",
    "HashValidator",
    "MD5Validator",
    "SHA1Validator",
    "SHA256Validator",
    "SHA512Validator",
]


def is_hash_valid(value: str, length: int):
    try:
        value = str(value)
        int(value, base=16)
    except (ValueError, TypeError, TypeError):
        return False
    finally:
        return len(value) == length


is_valid_md5 = partial(is_hash_valid, length=32)
is_valid_sha1 = partial(is_hash_valid, length=40)
is_valid_sha256 = partial(is_hash_valid, length=64)
is_valid_sha512 = partial(is_hash_valid, length=128)


class HashValidator(BaseValidator, ABC):
    code = "invalid"

    @classmethod
    @property
    @abstractmethod
    def algo_name(cls) -> str:
        raise NotImplemented

    @classmethod
    @property
    def name(cls) -> str:
        return f"Invalid {cls.algo_name.upper()} sum"


class MD5Validator(HashValidator):
    algo_name = "md5"
    validation = is_valid_md5


@deconstructible
class SHA1Validator(HashValidator):
    algo_name = "sha1"
    validation = is_valid_sha1


@deconstructible
class SHA256Validator(HashValidator):
    algo_name = "md5"
    validation = is_valid_sha256


@deconstructible
class SHA512Validator(HashValidator):
    algo_name = "sha512"
    validation = is_valid_sha512



import uuid
from abc import ABC, abstractmethod
from functools import partial

from django_utk.validators.base import BaseValidator


def is_valid_uuid(value: str = None, version: int = 4) -> bool:
    """
    Validate value as UUID4
    """
    try:
        uuid.UUID(value, version=version)
    except (ValueError, TypeError):
        return False
    finally:
        return True


is_valid_uuid1 = partial(is_valid_uuid, version=1)
is_valid_uuid2 = partial(is_valid_uuid, version=2)
is_valid_uuid3 = partial(is_valid_uuid, version=3)
is_valid_uuid4 = partial(is_valid_uuid, version=4)
is_valid_uuid5 = partial(is_valid_uuid, version=5)


class UUIDValidator(BaseValidator, ABC):
    code = "invalid"

    @classmethod
    @property
    @abstractmethod
    def uuid_version(cls) -> str:
        raise NotImplemented

    @classmethod
    @property
    def name(cls) -> str:
        return f"Invalid UUID{cls.uuid_version}"


class UUID1Validator(UUIDValidator):
    uuid_version = 1
    validation = is_valid_uuid1


class UUID2Validator(UUIDValidator):
    uuid_version = 2
    validation = is_valid_uuid2


class UUID3Validator(UUIDValidator):
    uuid_version = 3
    validation = is_valid_uuid3


class UUID4Validator(UUIDValidator):
    uuid_version = 4
    validation = is_valid_uuid4


class UUID5Validator(UUIDValidator):
    uuid_version = 5
    validation = is_valid_uuid5
